// serialhack - send OSC messages from the serial port out over UDP
//  Joel Darnauer Wed Feb 02 04:26:29 2005

// Thanks to Kim Poulsen who posted the serial port read routines to the net.

#include <fcntl.h>
#include <termios.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

////////////////////////////////////////////////////////////////
// MP and OSC definitions
////////////////////////////////////////////////////////////////
#include "htmsocket.h"

typedef struct serial_port {
  char *name;
  int fd;
  int baud_rate; // not used
} serial_port_t;



void usage () {
  fprintf(stderr, "Usage: serialosc2udp.exe [-v] <COM_port> <UDP_host> <UDP_port>\n");
  fprintf(stderr, "Example: serialosc2udp.exe -v COM2 localhost 9999\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "Listens for OSCserial messages from the AVR on a com port and sends\n"); 
  fprintf(stderr, "them to a specified UDP destination.\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "Requires cygwin1.dll <www.cygwin.org>\n");
  fprintf(stderr, "Bug: baud rate not working now - set within windows using hyperterminal first.\n");
  fprintf(stderr, "Bug: Serial port can close unexpectedly during high CPU load.\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "Contact joel for help with this program.\n");
  exit (-1);
}


#define OSC_HEADER ((char)0xBE)

void print_byte(unsigned char c) {
  //	  if (c > ' ' && c < '~') {
  //	    printf("%c", c);
  //	  }
  //	  else {
  //	    printf("(/%03d)", (unsigned int) c);
  //	  }
  printf("(0x%02x)", c);
}


// The defines in termios.h are meaningless for windows
dump_term_attr (int fd) {
  struct termios options;
  tcgetattr(fd, &options);

  printf ("options.c_iflag=	0x%08x\n", options.c_iflag);
  printf ("options.c_oflag=	0x%08x\n", options.c_oflag);
  printf ("options.c_cflag=	0x%08x\n", options.c_cflag);
  printf ("options.c_lflag=	0x%08x\n", options.c_lflag);
  printf ("options.c_line=	%c\n", options.c_line);
  printf ("options.c_ispeed=	0x08%d\n", options.c_ispeed);
  printf ("options.c_ospeed=	0x08%d\n", options.c_ospeed);

  printf ("ospeed = %d\n",  cfgetospeed(&options));

}



int open_serial_port (serial_port_t * portp) 
{
  if (portp->name) {
    portp->fd = open(portp->name, O_RDWR | O_NOCTTY );
    if (portp->fd < 0) {
      fprintf(stderr, "Couldn't open com port %s\n",  portp->name);
      return 0;
    }
    else {
      fprintf(stderr, "Com port open %s with file descriptor %d\n",  portp->name, portp->fd);

      if (0) {
	int baud_flags=0;
	//Make read blocking for now to avoid hogging CPU
	//fcntl(port.fd, F_SETFL, FNDELAY); // read returns immediately
	if (portp->baud_rate==57600)
	  baud_flags=B57600;
	else if (portp->baud_rate==9600)
	  baud_flags=B9600;
	else if (portp->baud_rate==38400)
	  baud_flags=B38400;
	else {
	  fprintf(stderr, "Error: baud rate of %d is not supported\n", portp->baud_rate);
	  exit(0);
	}
	fprintf(stdout, "Baud rate %d requested\n", portp->baud_rate);

	// Spy on terminal settings
	dump_term_attr(portp->fd);

	// Just assume that we can keep up with the programs
	struct termios options;
	options.c_ospeed &= ~CBAUDEX;      /* Mask the baud rate bits */
	options.c_ospeed |= baud_flags;    /* Set the new baud rate */

	tcsetattr(portp->fd, TCSANOW, &options);

	dump_term_attr(portp->fd); // no effect
      }

      return 1;
    }
  }
  else {
      fprintf(stderr, "Cant open unnamed serial port\n");
      return 0;
  }
}


// Get one character from the input device, reading a block of
// bytes if the local buffer for this file is empty.  We do this to
// avoid repeated system calls.

#define SERIAL_BUF_SIZE		10240

char read_char (serial_port_t * portp) 
{
  static char serial_in_buf[SERIAL_BUF_SIZE]; // buffer of characters
  static int  serial_in_pos=0;  // pointer to the last character produced by read_char
  static int  serial_in_lastchar=0; // pointer to the last valid character in the buffer

  if (serial_in_pos >= serial_in_lastchar) {
    // block waiting for more data
    int bytes=0;
    while (bytes <= 0) {
      bytes=read(portp->fd, serial_in_buf, SERIAL_BUF_SIZE);

      if (bytes == -1) {
	printf("Error trying to read serial port: (errno=%d (EAGAIN=%d, EWOULDBLOCK=%d, EACCES=%d))>...\n", 
	       errno, EAGAIN, EWOULDBLOCK, EACCES);
	// Something bad has happened while reading characters
	// This wont normally correct itself, to try to re-open the port
	close(portp->fd);
	open_serial_port(portp);
      }
      else {
#ifdef DEBUG
	printf("<read %d bytes>...\n", bytes);
#endif
      }
    }
    serial_in_lastchar=bytes-1;
    serial_in_pos=-1;
  }

  // Return the next available character
  return serial_in_buf[++serial_in_pos];
}


void wait_for_header (serial_port_t * portp) {

#ifdef DEBUG
    printf("\n\nwait for header...\n");
#endif
    // Synch until reading the first header byte

    while (1) {
      char byte=read_char(portp);
      if (byte==OSC_HEADER)
	break;
      else 
	print_byte(byte);
    }
}


// Assumes that the last char received was a header
// Takes the next character to be the size of the packet in bytes
// and reads the remaining characters from the data stream
// Checks the checksum last of all

int get_packet_body(serial_port_t *portp, char* packet_buf, int bufsize) {
  char read_buf[1];  // store chars coming in

  // Add the header to the packet
  packet_buf[0] = OSC_HEADER;

  // Read the size field
  packet_buf[1]=read_char(portp);
  int pkt_body_size=packet_buf[1];
#ifdef DEBUG
  printf("Got packet size %d\n", pkt_body_size);
#endif
  //print_byte(packet_buf[0]);
  //print_byte(packet_buf[1]);

  if (pkt_body_size % 4) {
    printf("ERROR: packet size %d is not a multiple of 4%d\n", pkt_body_size);    
    return 0;
  }

  // Packet size does not include header, size or checksum bytes
  int i;
  for (i=2; i<pkt_body_size+3; i++) {
    packet_buf[i]=read_char(portp);
    //print_byte(packet_buf[i]);
  }

  // Check checksum
  unsigned char cksum_rcvd=packet_buf[pkt_body_size+2];
#ifdef DEBUG
  printf("Received checksum=%d\n", cksum_rcvd);    
#endif

  unsigned char cksum_calc=0;
  for (i=2; i<pkt_body_size+2; i++) {
    cksum_calc += packet_buf[i];
  }
#ifdef DEBUG
  printf("Calculated checksum=%d\n", cksum_calc);    
#endif

  if (cksum_calc != cksum_rcvd) {
    printf("ERROR: packet cksum %d does not match computed cksum of %d\n", 
	   cksum_rcvd, cksum_calc);
    return 0;
  }
  else {
    return pkt_body_size+3;
  }
}


void dump_packet (int packet_size, char * packet_buf) {
  int i;
  printf("Packet=[");
  for (i=0; i<packet_size; i++) {
    print_byte(packet_buf[i]);
  }
  printf("]\n\n");
}



int main(int ARGC, char* ARGV[])
{
  serial_port_t port;
  serial_port_t *portp = &port;

  ////////////////////////////////////////////////////////////////
  // Check and process args
  ////////////////////////////////////////////////////////////////
  if (ARGC != 4 && ARGC != 5) {
    printf("ARGC=%d", ARGC);
    usage();
  }

  char * program_name = ARGV[0];
  int verbose=0;
  
  if (ARGV[1][0] == '-' && ARGV[1][1] == 'v') {
    printf("Verbose mode\n");
    verbose=1;
    // Shift args
    ARGV[1]=ARGV[2];
    ARGV[2]=ARGV[3];
    ARGV[3]=ARGV[4];
    ARGV[4]=0;
  }

  char * com_port = ARGV[1];
  char * udp_host = ARGV[2];
  char * udp_port_str = ARGV[3];
  //  char * baud_str = ARGV[2];


  printf("Running serialhack with args: com=%s, udp_host=%s, port=%s\n", 
	 com_port, udp_host, udp_port_str);

  int udp_port_num;
  if (sscanf(udp_port_str, "%d", &udp_port_num)!=1) {
    fprintf(stderr, "Error: <%s> is not a valid udp port number\n", udp_port_str);
    usage();
  }
  
  //  int baud_rate;
  //  if (sscanf(baud_str, "%d", &baud_rate)!=1) {
  //    fprintf(stderr, "Error: <%s> is not a valid baud rate\n", baud_str);
  //    usage();
  //}

  ////////////////////////////////////////////////////////////////
  // Open UDP Port
  ////////////////////////////////////////////////////////////////
  void * udp_socket = OpenHTMSocket(udp_host, udp_port_num);
  if (!udp_socket) {
    fprintf(stderr, "Couldn't open UDP socket to %s port %d\n", 
	    udp_host, udp_port_num);
    exit(-1);
  }

  ////////////////////////////////////////////////////////////////
  // Open serial port
  ////////////////////////////////////////////////////////////////
  {
    portp->name = com_port; // a little dangerous, since com_port is on the stack for main
    //    portp->baud_rate= baud_rate;

    if (!open_serial_port(portp)) {
      perror(com_port);
      fprintf(stderr, "Couldn't open com port %s\n",  com_port);
      exit(-1); 
    }
  }


  ////////////////////////////////////////////////////////////////
  // Listen to the serial port and send 
  ////////////////////////////////////////////////////////////////
  {

#define BUFSIZE 2000
    char packet_buf[BUFSIZE];
    int i;

    while (1) {
      wait_for_header(portp);
      int packet_size;
      if (packet_size=get_packet_body(portp, packet_buf, BUFSIZE)) {
	if (verbose) 
	  dump_packet(packet_size, packet_buf);

	// Send only the actual message over the udp connection
	// The checksum, header and size bytes are not part of the OSC spec.

	char * udp_packet_buf = packet_buf+2;
	int udp_packet_size = packet_size-3;
	SendHTMSocket(udp_socket, udp_packet_size, udp_packet_buf);
#if DEBUG
	printf("Sending packet size=%d\n", packet_size);
#endif
      }
      else {
	printf("ERROR: dropping corrupted packet(s)...\n");
      }
    }
  }
}
